import os
import sys
import pathlib
import hashlib

import argh
from argh.decorators import named

from Bio import SeqIO
from Bio.Seq import Seq

from konstel import __version__
from konstel.res import alphabets
import konstel.schema as schema 

import konstel.classes as classes
import konstel.formats as formats
import konstel.encodings as encodings
import konstel.helpers as helpers


def prepare(string, spec):
    ''''''
    if spec['remove_whitespace']:
        string = string.translate(str.maketrans('', '', ' \n\t\r'))
    if any(spec['strip_characters']):  # any() since schema default [''] is truthy
        string = string.strip(''.join(spec['strip_characters']))
    return string


def validate(string, spec):
    ''''''
    if spec['min_length']:
        if len(string) < spec['min_length']:
            raise RuntimeError(f'Validation failed: min_length')
    if spec['max_length']:
        if len(string) > spec['max_length']:
            raise RuntimeError(f'Validation failed: max_length')
    return True


def run_directive(string, scheme, directive, spec):
    ''''''
    if 'prepare' in spec[scheme]['directives'][directive]:
        string = prepare(string, spec[scheme]['directives'][directive]['prepare'])
    if 'validate' in spec[scheme]['directives'][directive]:
        validate(string, spec[scheme]['directives'][directive]['validate'])
    if spec[scheme]['directives'][directive]['helper']:
        helper_name = f"{scheme}_{directive}".replace('-','_')
        helper = getattr(helpers, helper_name)
        string = helper(string)
    return string


def generate_hash(string, algorithm):
    ''''''
    string_hash = getattr(hashlib, algorithm)(string.encode())
    return string_hash


def generate_output(string_hash, spec, no_prefix):
    ''''''
    encodings_raw = {n: getattr(encodings, m['type'])(string_hash) for n, m in spec.items()}
    encodings_fmt = {}
    for name, encoding_raw in encodings_raw.items():
        prefix = spec[name]['prefix'] if not no_prefix else ''
        length = spec[name]['length'] if 'length' in spec[name] else len(encoding_raw)
        encodings_fmt[name] = f"{prefix}{encoding_raw[:length]}"
        if spec[name].get('include_full'):
            encodings_fmt[f'{name}-full'] = f"{prefix}{encoding_raw}"
    return encodings_fmt


def format_output(outputs, output_type):
    '''Returns string format'''
    if output_type == 'dict':
        return outputs
    elif output_type == 'tab':
        outputs_fmt = ''
        for k, v in outputs.items():
            outputs_fmt += f'{k}\t{v}\n'
        return outputs_fmt
    elif output_type == 'table':
        outputs_fmt = ''
        for k, v in outputs.items():
            outputs_fmt += f'{k:<20} {v:<15}\n'
        return outputs_fmt

@named('scheme')
def generate_scheme(
        scheme: 'scheme name; specify {scheme}.{directive} if multiple directives are defined',
        string: 'input string' = '',
        file: 'input file path' = '',
        format: 'input format; mandatory if more than one format in scheme' = '',
        output: 'output format' = 'dict',
        no_prefix: 'hide encoding prefix; overrides scheme' = False):
    '''Generate identifier(s) for input string or file path according to specified scheme'''
    PACKAGE_PATH = os.path.dirname(os.path.dirname(__file__))
    scheme, _, directive = scheme.partition('.')
    print(f'Using scheme {scheme} ({directive})', file=sys.stderr)

    # Load scheme specification
    yaml_path = pathlib.Path(f'{PACKAGE_PATH}/schemes/{scheme}.yaml')
    if not os.path.exists(yaml_path):
        raise FileNotFoundError(f'Failed to open specification for scheme {scheme}')
    spec = schema.load_scheme(yaml_path.read_text()).data

    # Validate presence and unambiguity of scheme, directive and format
    if scheme not in spec:
        raise RuntimeError(f'Unrecognised scheme {scheme}.')
    if not directive:
        if len(spec[scheme]['directives']) == 1:  # One option
            directive = next(iter(spec[scheme]['directives']))
        else:
            raise RuntimeError(f"Ambiguous directive for scheme {scheme}. "
                               f"Options: {', '.join(spec[scheme]['directives'].keys())}")
    if directive not in spec[scheme]['directives']:
        raise RuntimeError(f'Unrecognised directive {directive} for scheme {scheme}')
    if not format:
        if len(spec[scheme]['directives'][directive]['formats']) == 1:
            format = next(iter(spec[scheme]['directives'][directive]['formats']))
        else:
            raise RuntimeError(f'Ambiguous format for directive {directive} of scheme {scheme}')
    
    # Validate string and file arguments, read file contents into string
    if string and file:
        raise RuntimeError(f'Specify either a string or a file path')
    if not string and not file:
        raise RuntimeError(f'Unspecified string or file path')
    if file and not os.path.exists(file):
        raise FileNotFoundError(f'File {file} not found ')
    if file:
        string = pathlib.Path(file).read_text()
    string = getattr(formats, format)(string)

    # Validate output
    if not output in schema.OUTPUT_TYPES:
        raise RuntimeError(f'Unrecognised output type {output}. Options: {schema.OUTPUT_TYPES}')

    # Handle DAG of directives
    dag = [directive]
    target = spec[scheme]['directives'][directive].get('target')
    if target:
        g = {s: spec[scheme]['directives'][s].get('target') for s in spec[scheme]['directives']}
        d = target
        while d:
            dag.append(d)
            d = g[d]
    for d in dag:
        string = run_directive(string, scheme, d, spec)
    string_hash = generate_hash(string, spec[scheme]['algorithm'])
    outputs = generate_output(string_hash, spec[scheme]['encodings'], no_prefix)

    return format_output(outputs, output)


# def validate_scheme():
#     if hash_algorithm not in hash_funcs:
#         raise RuntimeError(f'Unrecognised hash function {hash_algorithm}')

# Validate target graph
# Can't have both helper and hash function


def main():
    argh.dispatch_commands([generate_scheme])



import base64


def base32(string_hash):
    '''Returns lowercased and padding-stripped base32 encoding of a hashlib Hash'''
    return base64.b32encode(string_hash.digest()).decode().lower().rstrip('=')


def cbase32(string_hash):
    '''Returns lowercased cbase32 encoding of a hashlib Hash instance'''
    hash_b32 = base32(string_hash)
    base32_symbols = 'abcdefghijklmnopqrstuvwxyz234567'
    cbase32_symbols = '0123456789abcdefghjkmnpqrstvwxyz'
    base32_to_cbase32 = dict(zip(base32_symbols, cbase32_symbols))
    return ''.join([base32_to_cbase32[c] for c in hash_b32])


def phonemes_10_5(string_hash):
    '''
    Returns word comprising consonant-vowel phonemes from hashlib Hash
    Maps 10 consonants and 5 vowels to base10 encoded hash
    2.8 bits per character
    '''
    hash_b10 = str(int(string_hash.hexdigest(), 16))
    vowel_map = dict(zip(map(str, range(10)), 'aeiou'*2))
    consonant_map = dict(zip(map(str, range(10)), 'fhklmprstv'))
    word = ''
    for i, char in enumerate(hash_b10):
        if not i % 2:
            word += consonant_map[char]
        else:
            word += vowel_map[char]
    return word


def phonemes_16_4(string_hash):
    '''
    Returns word comprising consonant-vowel phonemes from hashlib Hash
    Maps 16 consonants and 4 vowels to six bit windows of hash
    3 bits per character
    '''
    hash_b2 = bin(int(string_hash.hexdigest(), 16))[2:]  # Binary string from hex string
    vowel_map = dict(zip(range(4), 'aiou'))
    consonant_map = dict(zip(range(16), 'bdfghjklmnprstvz'))
    word = ''
    for pos in range(0, len(hash_b2), 6):
        window = hash_b2[pos:pos+6]
        if len(window) < 6:  # Pad with trailing zeros
            window = f'{window:<06}'
        c, v = int(window[:4], 2), int(window[4:], 2)
        word += f'{consonant_map[c]}{vowel_map[v]}'
    return word


def phonemes_16_4_bits(string_hash, result_len=None):
    '''
    Returns word comprising consonant-vowel phonemes from hashlib Hash
    Maps 16 consonants and 4 vowels to six bit windows of hash
    3 bits per character
    '''
    import itertools
    from math import log2
    phonemes = [''.join(i) for i in itertools.product('bdfghjklmnprstvz', 'aiou')]
    phoneme_bit_size = log2(len(phonemes))
    assert phoneme_bit_size.is_integer(), 'Must have a power of two number of phonemes'

    hash_b2 = int(string_hash.hexdigest(), 16)

    if result_len is None:
        result_len = int(8 * string_hash.digest_size / phoneme_bit_size)
    front_offset = int(8 * string_hash.digest_size - phoneme_bit_size)
    mask = (2 ** int(phoneme_bit_size) - 1) << front_offset

    word = ''
    for _ in range(result_len):
        phoneme_num = (hash_b2 & mask) >> front_offset
        word += phonemes[phoneme_num]
        hash_b2 = hash_b2 << int(phoneme_bit_size)
    return word


import io

from Bio import SeqIO


def fasta(string):
	string_as_file = io.StringIO(string)
	record = SeqIO.read(string_as_file, 'fasta')
	assert len(record.seq) >= 1, 'Sequence is empty'
	return str(record.seq).upper()



import hashlib
import inspect

from strictyaml import load, Bool, Int, Str, Seq, Map, Enum, MapPattern, Optional, YAMLValidationError

from konstel import formats
from konstel import encodings


ALGORITHMS = hashlib.algorithms_available
FORMATS = [o[0] for o in inspect.getmembers(formats, inspect.isfunction)]
ENCODINGS = [o[0] for o in inspect.getmembers(encodings, inspect.isfunction)]
OUTPUT_TYPES = {'dict', 'tab', 'table'}


# For when classes are implemented
# classes = [obj for name, obj in inspect.getmembers(sys.modules[__name__], inspect.isclass)
#           if obj.__module__ is __name__]
# [m[0] for m in inspect.getmembers(my_module, inspect.isclass) if m[1].__module__ == 'my_module']


def load_scheme(yaml_text):
    '''
    Some optional keys have enforced default values, otherwise use dict.get()
    '''
    schema = MapPattern(
        Str(), Map({
            'description': Str(),
            Optional('alias'): Str(),
            'version': Str(),
            'directives': MapPattern(
                Str(), Map({
                    'description': Str(),
                    'class': Str(),
                    'formats': Seq(Enum(FORMATS)),
                    Optional('prepare'): Map({
                        Optional('remove_whitespace', default=False): Bool(),
                        Optional('strip_characters', default=['']): Seq(Str()),
                    }),
                    Optional('validate'): Map({
                        Optional('min_length'): Int(),
                        Optional('max_length'): Int(),
                    }),
                    Optional('target'): Str(),
                    Optional('helper', default=False): Bool()
                }),
            ),
            'algorithm': Enum(ALGORITHMS),
            'encodings': MapPattern(
                Str(), Map({
                    'type': Enum(ENCODINGS),
                    'length': Int(),
                    Optional('prefix', default=''): Str(),
                    Optional('include_full', default=False): Bool(),
                    Optional('function', default=False): Bool(),
                })
            )
        })
    )
    return load(yaml_text, schema)



from Bio import SeqIO
from Bio.Seq import Seq


def sars_cov_2_s_genome(nuc_sequence):
    '''Returns translated SARS-CoV-2 spike sequence contained in nucleotide string'''
    nuc_sequence_fmt = nuc_sequence.translate(str.maketrans('', '', ' \n\t\r'))
    seq = Seq(nuc_sequence_fmt)
    s_start_seq = 'atgtttgttttt'  # First 4 codons of Wuhan-Hu-1
    s_end_seq = 'ttacattacacataa'  # Last 5 codons of Wuhan-Hu-1
    try:
        s_start_pos = str(seq).lower().index(s_start_seq)
        s_end_pos = str(seq).lower().index(s_end_seq) + len(s_end_seq)
        spike_nucl = seq[s_start_pos:s_end_pos]
        spike_prot = str(spike_nucl.ungap().translate()).strip('*')
    except Exception as e:
        raise RuntimeError('Problem extracting spike sequence').with_traceback(e.__traceback__)
    return spike_prot

def sars_cov_2_s_legacy_genome(nuc_sequence):
    return sars_cov_2_s_genome(nuc_sequence)


import konstel

from setuptools import setup


setup(name = 'konstel',
      version = konstel.__version__,
      description = 'Hash-based phonemic nomenclature schemes for biological sequences',
      url = 'http://github.com/bede/konstel',
      author = 'Bede Constantinides',
      author_email = 'bedeabc@gmail.com',
      license = 'LICENSE',
      packages=['konstel'],
      python_requires='>=3.6',
      install_requires=['argh', 'biopython', 'strictyaml'],
      entry_points = {'console_scripts':['konstel=konstel.konstel:main']},
      classifiers=['Intended Audience :: Science/Research',
                   'License :: OSI Approved :: MIT License',
                   'Natural Language :: English'])
